import _extends from "@babel/runtime/helpers/extends";
import _objectWithoutProperties from "@babel/runtime/helpers/objectWithoutProperties";
import cn from 'classnames';
import shallowEqual from 'fbjs/lib/shallowEqual';
import * as React from 'react';
import { accessibilityOverscanIndicesGetter, Grid } from "react-virtualized/dist/es/Grid";
import defaultRowRenderer from './defaultRowRenderer';
import { defaultControlStyle, defaultRowStyle, NodeRecord, Update } from './utils';

/**
 * This component renders a tree of elements using powerful virtualization technology to render
 * only visible elements to the HTML document.
 */
export default class Tree extends React.Component {
  constructor(...args) {
    var _temp;

    return _temp = super(...args), this.state = {
      order: [],
      useDynamicRowHeight: false
    }, this.grid = void 0, this.registry = {}, this.getRowHeight = ({
      index
    }) => {
      const {
        rowHeight
      } = this.props;
      const {
        order
      } = this.state;
      const id = order[index];
      const {
        node
      } = this.registry[id];
      return node.height || rowHeight;
    }, this.cellRenderer = ({
      rowIndex,
      style,
      isScrolling,
      key
    }) => {
      const {
        controlClassName,
        controlStyle,
        rowClassName,
        nodeNestingMultiplier,
        onRowClick,
        onRowDoubleClick,
        onRowMouseOver,
        onRowMouseOut,
        onRowRightClick,
        rowHeight,
        rowRenderer,
        rowStyle
      } = this.props;
      const {
        order
      } = this.state; // TRICKY The style object is sometimes cached by Grid.
      // This prevents new style objects from bypassing shallowCompare().
      // However as of React 16, style props are auto-frozen (at least in dev mode)
      // Check to make sure we can still modify the style before proceeding.
      // https://github.com/facebook/react/commit/977357765b44af8ff0cfea327866861073095c12#commitcomment-20648713

      const {
        writable
      } = Object.getOwnPropertyDescriptor(style, 'width');

      if (writable) {
        // By default, Tree cells should be 100% width.
        // This prevents them from flowing under a scrollbar (if present).
        style.width = '100%';
      }

      const id = order[rowIndex];
      const record = this.registry[id];
      const {
        node: {
          childrenCount,
          nestingLevel,
          nodeData,
          height,
          style: specificRowStyle
        },
        isOpened,
        onNodeToggle
      } = record;
      return rowRenderer({
        childrenCount,
        className: rowClassName,
        controlClassName,
        controlStyle: Object.assign({}, defaultControlStyle, controlStyle),
        id,
        index: rowIndex,
        isOpened,
        isScrolling,
        key,
        nestingLevel,
        nodeData,
        onNodeToggle,
        onRowClick,
        onRowDoubleClick,
        onRowMouseOut,
        onRowMouseOver,
        onRowRightClick,
        style: Object.assign({}, style, defaultRowStyle, rowStyle, specificRowStyle, {
          height: height || rowHeight,
          marginLeft: nestingLevel * nodeNestingMultiplier,
          overflow: 'hidden'
        })
      });
    }, this.finishNodeToggling = async () => {
      await this.recomputeTree(Update.Nodes);
    }, this.onScroll = ({
      clientHeight,
      scrollHeight,
      scrollTop
    }) => {
      const {
        onScroll
      } = this.props;
      onScroll({
        clientHeight,
        scrollHeight,
        scrollTop
      });
    }, this.onSectionRendered = ({
      rowOverscanStartIndex,
      rowOverscanStopIndex,
      rowStartIndex,
      rowStopIndex
    }) => {
      const {
        onRowsRendered
      } = this.props;
      onRowsRendered({
        overscanStartIndex: rowOverscanStartIndex,
        overscanStopIndex: rowOverscanStopIndex,
        startIndex: rowStartIndex,
        stopIndex: rowStopIndex
      });
    }, this.setRef = grid => {
      this.grid = grid;
    }, _temp;
  }

  async componentWillMount() {
    await this.recomputeTree(Update.NodesAndOpenness);
  }

  async componentWillReceiveProps({
    update
  }) {
    if (update !== Update.None) {
      await this.recomputeTree(update);
    }
  }

  forceUpdateGrid(callback) {
    if (this.grid) {
      this.grid.forceUpdate(callback);
    }
  }
  /** See Grid#getOffsetForCell */


  getOffsetForRow({
    alignment,
    index
  }) {
    if (this.grid) {
      const {
        scrollTop
      } = this.grid.getOffsetForCell({
        alignment,
        columnIndex: 0,
        rowIndex: index
      });
      return scrollTop;
    }

    return 0;
  }
  /** CellMeasurer compatibility */


  invalidateCellSizeAfterRender({
    columnIndex,
    rowIndex
  }) {
    if (this.grid) {
      this.grid.invalidateCellSizeAfterRender({
        columnIndex,
        rowIndex
      });
    }
  }
  /** See Grid#measureAllCells */


  measureAllRows() {
    if (this.grid) {
      this.grid.measureAllCells();
    }
  }
  /** CellMeasurer compatibility */


  recomputeGridSize({
    columnIndex = 0,
    rowIndex = 0
  } = {}) {
    if (this.grid) {
      this.grid.recomputeGridSize({
        columnIndex,
        rowIndex
      });
    }
  }
  /**
   * Recomputes Tree component basing on update type.
   *
   * Method calls "nodeGetter" internally and flattens tree structure to an array. Depending on "update" value it runs
   * one of the following algorithms:
   * 1) Update.NodesAndOpenness. Requires full node metadata. Updates order, number and rendering data of all
   * nodes. Overrides current openness state with "openedByDefault" value.
   * 2) Update.Nodes. Requires full node metadata. Updates order, number and rendering data of all nodes. Preserves
   * openness state of all nodes.
   * 3) Update.Order. Requires only node id. Updates nodes order (useful for sorting etc.). Preserves openness
   * state of all nodes.
   * 4) Update.None. Updates nothing.
   *
   * @param update
   *
   * @returns promise resolves when re-rendering is ended
   */


  async recomputeTree(update) {
    if (update === Update.None) {
      return void 0;
    }

    const {
      nodeGetter,
      rowHeight
    } = this.props;
    const order = [];
    let useDynamicRowHeight = false;
    const refresh = update === Update.Nodes || update === Update.NodesAndOpenness;
    const ignoreInnerState = update === Update.NodesAndOpenness;
    const g = nodeGetter(refresh);
    let isPreviousOpened = false; // tslint:disable-next-line:no-constant-condition

    while (true) {
      const {
        value,
        done
      } = g.next(isPreviousOpened);

      if (done || !value) {
        break;
      }

      if (typeof value === 'string') {
        order.push(value);
        isPreviousOpened = this.registry[value].isOpened;
      } else {
        const {
          height,
          id,
          isOpenedByDefault
        } = value;
        const record = this.registry[id];

        if (!record) {
          this.registry[id] = new NodeRecord(value, isOpenedByDefault, this.finishNodeToggling);
        } else {
          record.node = value;

          if (ignoreInnerState) {
            record.isOpened = isOpenedByDefault;
          }
        }

        order.push(id);
        isPreviousOpened = this.registry[id].isOpened;

        if (height && height !== rowHeight) {
          useDynamicRowHeight = true;
        }
      }
    }

    return new Promise(resolve => {
      this.setState({
        order,
        useDynamicRowHeight
      }, () => {
        if (useDynamicRowHeight) {
          this.recomputeGridSize();
          resolve();
        } else {
          this.forceUpdateGrid(resolve);
        }
      });
    });
  }

  render() {
    const _props = this.props,
          {
      className,
      noRowsRenderer,
      onScroll,
      rowHeight,
      scrollToIndex,
      update,
      width // tslint:disable-line:trailing-comma

    } = _props,
          other = _objectWithoutProperties(_props, ["className", "noRowsRenderer", "onScroll", "rowHeight", "scrollToIndex", "update", "width"]);

    const {
      order,
      useDynamicRowHeight
    } = this.state;
    const classNames = cn('ReactVirtualized__Tree', className);
    return React.createElement(Grid, _extends({
      autoContainerWidth: true,
      cellRenderer: this.cellRenderer,
      className: classNames,
      columnWidth: width,
      columnCount: 1,
      noContentRenderer: noRowsRenderer,
      onScroll: this.onScroll,
      onSectionRendered: this.onSectionRendered,
      ref: this.setRef,
      rowCount: order.length,
      rowHeight: useDynamicRowHeight ? this.getRowHeight : rowHeight,
      scrollToRow: scrollToIndex,
      width: width
    }, other));
  }
  /** See Grid#scrollToPosition */


  scrollToPosition(scrollTop = 0) {
    if (this.grid) {
      this.grid.scrollToPosition({
        scrollTop
      });
    }
  }
  /** See Grid#scrollToCell */


  scrollToRow(index = 0) {
    if (this.grid) {
      this.grid.scrollToCell({
        columnIndex: 0,
        rowIndex: index
      });
    }
  }

  shouldComponentUpdate(_ref, nextState) {
    let {
      update
    } = _ref,
        nextOther = _objectWithoutProperties(_ref, ["update"]);

    const _props2 = this.props,
          {
      update: _
    } = _props2,
          other = _objectWithoutProperties(_props2, ["update"]);

    return !shallowEqual(other, nextOther) || !shallowEqual(this.state, nextState) || update !== Update.None;
  }
  /**
   * Overrides openness state of nodes width ids specified as map keys to map values respectively.
   * @param map
   */


  async toggleNodes(map) {
    // tslint:disable-next-line:forin no-for-in
    for (const id in map) {
      this.registry[id].isOpened = map[id];
    }

    await this.recomputeTree(Update.Nodes);
  }

}
Tree.defaultProps = {
  autoHeight: false,
  estimatedRowSize: 30,
  noRowsRenderer: () => null,
  nodeNestingMultiplier: 10,
  onRowsRendered: () => undefined,
  onScroll: () => undefined,
  overscanIndicesGetter: accessibilityOverscanIndicesGetter,
  overscanRowCount: 10,
  rowRenderer: defaultRowRenderer,
  scrollToAlignment: 'auto',
  scrollToIndex: -1,
  style: {},
  update: Update.None
};